import logging
import os
import xml.etree.ElementTree as ET
from datetime import datetime
from typing import Optional, Union

from ..gpx_elements import (
    Bounds,
    Copyright,
    Email,
    Extensions,
    Gpx,
    Link,
    Metadata,
    Person,
    Point,
    PointSegment,
    Route,
    Track,
    TrackSegment,
    WayPoint,
)
from .parser import POSSIBLE_TIME_FORMATS
from .xml_parser import XMLParser


class GPXParser(XMLParser):
    """
    GPX file parser.
    """

    def __init__(
        self,
        file_path: Optional[str] = None,
        xml_schemas: bool = True,
        xml_extensions_schemas: bool = False,
    ) -> None:
        """
        Initialize GPXParser instance.

        Parameters
        ----------
        file_path : Optional[str], optional
            Path to the file to parse, by default None
        xml_schemas : bool, optional
            Toggle schema verification during parsing, by default True
        xml_extensions_schemas : bool, optional
            Toggle extensions schema verificaton durign parsing.
            Requires internet connection and is not guaranted
            to work., by default False
        """
        if not file_path.endswith(".gpx"):
            return

        super().__init__(file_path, xml_schemas, xml_extensions_schemas)

        if self.file_path is not None and os.path.exists(self.file_path):
            self.parse()
        else:
            logging.warning("File path does not exist")

    def _find_precisions(self):
        """
        Find decimal precision of any type of value in a GPX file (latitude, elevation...).
        Also find if the GPX file contains elevation data.
        """
        # Point
        track = self.xml_root.findall("trk", self.name_spaces)[0]
        segment = track.findall("trkseg", self.name_spaces)[0]
        point = segment.findall("trkpt", self.name_spaces)[0]

        ele_text = point.findtext("ele", namespaces=self.name_spaces)
        if ele_text is not None:
            self.ele_data = True
        else:
            self.ele_data = False

        self.precisions["lat_lon"] = self.find_precision(point.get("lat"))
        self.precisions["elevation"] = self.find_precision(ele_text)

    def _find_time_element(self) -> Union[str, None]:
        """
        Find a time element in GPX file.

        Returns:
            Union[str, None]: Time element.
        """
        # Use time from metadata
        metadata = self.xml_root.find("metadata", self.name_spaces)
        if metadata is not None:
            time_str = metadata.findtext("time", namespaces=self.name_spaces)
            if time_str is not None:
                return time_str

        # Use time from track point
        track = self.xml_root.findall("trk", self.name_spaces)[
            0
        ]  # Optimise, load only once??
        segment = track.findall("trkseg", self.name_spaces)[0]
        point = segment.findall("trkpt", self.name_spaces)[0]
        time_str = point.findtext("time", namespaces=self.name_spaces)
        if time_str is not None:
            return time_str

        # No time element at all...
        return None

    def _find_time_format(self):
        """
        Find the time format used in GPX file.
        Also find if the GPX file contains time data.
        """
        time_str = self._find_time_element()
        if time_str is None:
            self.time_data = False
            logging.warning("No time element in GPX file.")
            return

        self.time_data = True
        for tf in POSSIBLE_TIME_FORMATS:
            try:
                datetime.strptime(time_str, tf)
                self.time_format = tf
                break
            except ValueError:
                pass
        else:
            logging.info(
                "Unknown time format. "
                "Default time format will be used uppon writting."
            )

    def _parse_bounds(self, bounds, tag: str = "bounds") -> Union[Bounds, None]:
        """
        Parse boundsType element from GPX file.

        Args:
            bounds (xml.etree.ElementTree.Element): Parsed bounds element.
            tag (str, Optional): XML tag. Defaults to "bounds".

        Returns:
            Bounds: Bounds instance.
        """
        if bounds is None:
            return None

        minlat = bounds.get("minlat")
        minlon = bounds.get("minlon")
        maxlat = bounds.get("maxlat")
        maxlon = bounds.get("maxlon")

        return Bounds(tag, minlat, minlon, maxlat, maxlon)

    def _parse_copyright(
        self, copyright_, tag: str = "copyright"
    ) -> Union[Copyright, None]:
        """
        Parse copyrightType element from GPX file.

        Args:
            copyright (xml.etree.ElementTree.Element): Parsed copyright element.
            tag (str, Optional): XML tag. Defaults to "copyright".

        Returns:
            Copyright: Copyright instance.
        """
        if copyright_ is None:
            return None

        author = copyright_.get("author")
        year = copyright_.findtext("year", namespaces=self.name_spaces)
        licence = copyright_.findtext("licence", namespaces=self.name_spaces)

        return Copyright(tag, author, year, licence)

    def _parse_email(self, email, tag: str = "email") -> Union[Email, None]:
        """
        Parse emailType element from GPX file.

        Args:
            email (xml.etree.ElementTree.Element): Parsed email element.
            tag (str, Optional): XML tag. Defaults to "email".

        Returns:
            Email: Email instance.
        """
        if email is None:
            return None

        id_ = email.get("id")
        domain = email.get("domain")

        return Email(tag, id_, domain)

    def _parse_extensions(
        self, extensions, element_type: str, tag: str = "extensions"
    ) -> Union[Extensions, None]:
        """
        Parse extensionsType element from GPX file.

        Args:
            extensions (xml.etree.ElementTree.Element): Parsed extensions element.
            tag (str, Optional): XML tag. Defaults to "extensions".

        Returns:
            Extensions: Extensions instance.
        """
        if extensions is None:
            return None

        def construct_dict(e0):
            e1s = [e1 for e1 in e0.iter()][1:]
            if len(e1s) > 0:
                d = {"attrib": dict(e0.items()), "elmts": {}}
                for e1 in e1s:
                    d["elmts"][e1.tag] = construct_dict(e1)
                return d
            else:
                return {"attrib": {}, "elmts": e0.text}

        ext = [e for e in extensions.iter()][1]
        values = {ext.tag: {}}
        values[ext.tag] = construct_dict(ext)

        # Etensions fields are based on the first occurance of a type encountered in the file
        if self.extensions_fields.get(element_type) is None:
            self.extensions_fields[element_type] = values

        return Extensions(tag, values)

    def _parse_link(self, link, tag: str = "link") -> Union[Link, None]:
        """
        Parse linkType element from GPX file.

        Args:
            link (xml.etree.ElementTree.Element): Parsed link element.
            tag (str, Optional): XML tag. Defaults to "link".

        Returns:
            Link: Link instance.
        """
        if link is None:
            return None

        href = link.get("href")
        text = link.findtext("text", namespaces=self.name_spaces)
        type_ = link.findtext("type", namespaces=self.name_spaces)

        return Link(tag, href, text, type_)

    def _parse_metadata(self, metadata, tag: str = "metadata") -> Union[Metadata, None]:
        """
        Parse metadataType element from GPX file.

        Args:
            metadata (xml.etree.ElementTree.Element): Parsed metadata element.
            tag (str, Optional): XML tag. Defaults to "metadata".

        Returns:
            Metadata: Metadata instance.
        """
        if metadata is None:
            return None

        name = metadata.findtext("name", namespaces=self.name_spaces)
        desc = metadata.findtext("desc", namespaces=self.name_spaces)
        author = self._parse_person(metadata.find("author", self.name_spaces))
        copyright_ = self._parse_copyright(metadata.find("copyright", self.name_spaces))
        link = self._parse_link(metadata.find("link", self.name_spaces))
        time = self.find_time(metadata, "time")
        keywords = metadata.findtext("keywords", namespaces=self.name_spaces)
        bounds = self._parse_bounds(metadata.find("bounds", self.name_spaces))
        extensions = self._parse_extensions(
            metadata.find("extensions", self.name_spaces), tag
        )

        return Metadata(
            tag,
            name,
            desc,
            author,
            copyright_,
            link,
            time,
            keywords,
            bounds,
            extensions,
        )

    def _parse_person(self, person, tag: str = "person") -> Union[Person, None]:
        """
        Parse personType element from GPX file.

        Args:
            person (xml.etree.ElementTree.Element): Parsed person element.
            tag (str, Optional): XML tag. Defaults to "person".

        Returns:
            Person: Person instance.
        """
        if person is None:
            return None

        name = person.findtext("name", namespaces=self.name_spaces)
        email = self._parse_email(person.find("email", self.name_spaces))
        link = self._parse_link(person.find("link", self.name_spaces))

        return Person(tag, name, email, link)

    def _parse_point_segment(
        self, point_segment, tag: str = "ptseg"
    ) -> Union[PointSegment, None]:
        """
        Parse ptsegType element from GPX file.

        Parameters
        ----------
        point_segment : xml.etree.ElementTree.Element
            Parsed point segment element
        tag : str, optional
            XML tag, by default "ptseg"

        Returns
        -------
        Union[PointSegment, None]
            PointSegment instance
        """
        if point_segment is None:
            return None

        pt = [self._parse_point(p) for p in point_segment.findall("pt")]

        return PointSegment(tag, pt)

    def _parse_point(self, point, tag: str = "pt") -> Union[Point, None]:
        """
        Parse ptType element from GPX file.

        Args:
            point (xml.etree.ElementTree.Element): Parsed pt element.
            tag (str, Optional): XML tag. Defaults to "pt".

        Returns:
            Point: Point instance.
        """
        if point is None:
            return None

        lat = self.get_float(point, "lat")
        lon = self.get_float(point, "lon")
        ele = self.find_float(point, "ele")
        time = self.find_time(point, "time")

        return Point(tag, lat, lon, ele, time)

    def _parse_route(self, route, tag: str = "rte") -> Union[Route, None]:
        """
        Parse rteType element from GPX file.

        Args:
            route (xml.etree.ElementTree.Element): Parsed rte element.
            tag (str, Optional): XML tag. Defaults to "rte".

        Returns:
            Route: Route instance.
        """
        if route is None:
            return None

        name = route.findtext("name", namespaces=self.name_spaces)
        cmt = route.findtext("cmt", namespaces=self.name_spaces)
        desc = route.findtext("desc", namespaces=self.name_spaces)
        src = route.findtext("src", namespaces=self.name_spaces)
        link = self._parse_link(route.find("link", self.name_spaces))
        number = self.find_int(route, "number")
        type_ = route.findtext("type", namespaces=self.name_spaces)
        extensions = self._parse_extensions(
            route.find("extensions", self.name_spaces), tag
        )
        rtept = [
            self._parse_way_point(way_point)
            for way_point in route.findall("rtept", self.name_spaces)
        ]

        return Route(tag, name, cmt, desc, src, link, number, type_, extensions, rtept)

    def _parse_track_segment(
        self, track_segment, tag: str = "trkseg"
    ) -> Union[TrackSegment, None]:
        """
        Parse trksegType element from GPX file.

        Args:
            track_segment (xml.etree.ElementTree.Element): Parsed trkseg element.
            tag (str, Optional): XML tag. Defaults to "trkseg".

        Returns:
            TrackSegment: TrackSegment instance.
        """
        if track_segment is None:
            return None

        trkpt = [
            self._parse_way_point(track_point, "trkpt")
            for track_point in track_segment.findall("trkpt", self.name_spaces)
        ]
        extensions = self._parse_extensions(
            track_segment.find("extensions", self.name_spaces), tag
        )

        return TrackSegment(tag, trkpt, extensions)

    def _parse_track(self, track, tag: str = "trk") -> Union[Track, None]:
        """
        Parse trkType element from GPX file.

        Args:
            track (xml.etree.ElementTree.Element): Parsed trk element.
            tag (str, Optional): XML tag. Defaults to "trk".

        Returns:
            Track: Track instance.
        """
        if track is None:
            return None

        name = track.findtext("name", namespaces=self.name_spaces)
        cmt = track.findtext("cmt", namespaces=self.name_spaces)
        desc = track.findtext("desc", namespaces=self.name_spaces)
        src = track.findtext("src", namespaces=self.name_spaces)
        link = self._parse_link(track.find("link", self.name_spaces))
        number = self.find_int(track, "number")
        type_ = track.findtext("type", namespaces=self.name_spaces)
        extensions = self._parse_extensions(
            track.find("extensions", self.name_spaces), tag
        )
        trkseg = [
            self._parse_track_segment(segment)
            for segment in track.findall("trkseg", self.name_spaces)
        ]

        return Track(tag, name, cmt, desc, src, link, number, type_, extensions, trkseg)

    def _parse_way_point(self, way_point, tag: str = "wpt") -> Union[WayPoint, None]:
        """
        Parse wptType element from GPX file.

        Args:
            way_point (xml.etree.ElementTree.Element): Parsed wpt element.
            tag (str, Optional): XML tag. Defaults to "person".

        Returns:
            WayPoint: WayPoint instance.
        """
        if way_point is None:
            return None

        lat = self.get_float(way_point, "lat")
        lon = self.get_float(way_point, "lon")
        ele = self.find_float(way_point, "ele")
        time = self.find_time(way_point, "time")
        mag_var = self.find_float(way_point, "magvar")
        geo_id_height = self.find_float(way_point, "geoidheight")
        geo_id_height = self.find_float(way_point, "geoidheight")
        name = way_point.findtext("name", namespaces=self.name_spaces)
        cmt = way_point.findtext("cmt", namespaces=self.name_spaces)
        desc = way_point.findtext("desc", namespaces=self.name_spaces)
        src = way_point.findtext("src", namespaces=self.name_spaces)
        link = self._parse_link(way_point.find("link", self.name_spaces))
        sym = way_point.findtext("sym", namespaces=self.name_spaces)
        type_ = way_point.findtext("type", namespaces=self.name_spaces)
        fix = way_point.findtext("fix", namespaces=self.name_spaces)
        sat = self.find_int(way_point, "sat")
        hdop = self.find_float(way_point, "hdop")
        vdop = self.find_float(way_point, "vdop")
        pdop = self.find_float(way_point, "pdop")
        age_of_gps_data = self.find_float(way_point, "ageofgpsdata")
        dgpsid = self.find_float(way_point, "dgpsid")
        extensions = self._parse_extensions(
            way_point.find("extensions", self.name_spaces), tag
        )

        return WayPoint(
            tag,
            lat,
            lon,
            ele,
            time,
            mag_var,
            geo_id_height,
            name,
            cmt,
            desc,
            src,
            link,
            sym,
            type_,
            fix,
            sat,
            hdop,
            vdop,
            pdop,
            age_of_gps_data,
            dgpsid,
            extensions,
        )

    def _parse_root_properties(self):
        """
        Parse XML properties from GPX file.
        """
        self.gpx.creator = self.xml_root.attrib["creator"]
        self.gpx.version = self.xml_root.attrib["version"]
        schema_location = self.xml_root.get(
            "{http://www.w3.org/2001/XMLSchema-instance}schemaLocation"
        ).split(" ")
        self.gpx.xsi_schema_location = [x for x in schema_location if x != ""]

    def _parse_root_metadata(self):
        """
        Parse metadataType elements from GPX file.
        """
        self.gpx.metadata = self._parse_metadata(
            self.xml_root.find("metadata", self.name_spaces)
        )

    def _parse_root_way_points(self):
        """
        Parse wptType elements from GPX file.
        """
        way_points = self.xml_root.findall("wpt", self.name_spaces)
        for way_point in way_points:
            self.gpx.wpt.append(self._parse_way_point(way_point))

    def _parse_root_routes(self):
        """
        Parse rteType elements from GPX file
        """
        routes = self.xml_root.findall("rte", self.name_spaces)
        for route in routes:
            self.gpx.rte.append(self._parse_route(route))

    def _parse_root_tracks(self):
        """
        Parse trkType elements from GPX file.
        """
        tracks = self.xml_root.findall("trk", self.name_spaces)
        for track in tracks:
            self.gpx.trk.append(self._parse_track(track))

    def _parse_root_extensions(self):
        """
        Parse extensionsType elements from GPX file.
        """
        extensions = self.xml_root.find("extensions", self.name_spaces)
        self.gpx.extensions = self._parse_extensions(extensions, "gpx")

    def parse(self) -> Gpx:
        """
        Parse GPX file.

        Returns:
            Gpx: Gpx instance.
        """
        # Parse GPX file
        try:
            self.xml_tree = ET.parse(self.file_path)
            self.xml_root = self.xml_tree.getroot()
        except Exception as err:
            logging.exception(
                "Unexpected %s, %s.\n" "Unable to parse GPX file.", err, type(err)
            )
            raise

        # Parse properties
        try:
            self._parse_root_properties()
        except:
            logging.error("Unable to parse properties in GPX file.")
            raise

        # Check XML schemas
        self.check_xml_schemas()

        # Find precisions
        self._find_precisions()

        # Find time format
        self._find_time_format()

        # Parse metadata
        try:
            self._parse_root_metadata()
        except:
            logging.error("Unable to parse metadata in GPX file.")
            raise

        # Parse way points
        try:
            self._parse_root_way_points()
        except:
            logging.error("Unable to parse way_points in GPX file.")
            raise

        # Parse routes
        try:
            self._parse_root_routes()
        except:
            logging.error("Unable to parse routes in GPX file.")
            raise

        # Parse tracks
        try:
            self._parse_root_tracks()
        except:
            logging.error("Unable to parse tracks in GPX file.")
            raise

        # Parse extensions
        try:
            self._parse_root_extensions()
        except:
            logging.error("Unable to parse extensions in GPX file.")
            raise

        logging.debug("Parsing complete!!")
        return self.gpx



import logging
import math as m

# latitude/longitude in GPX files is always in WGS84 datum
# WGS84 defined the Earth semi-major axis with 6378.137 km
# https://en.wikipedia.org/wiki/World_Geodetic_System#WGS84
EARTH_RADIUS = 6378.137 * 1000


def haversine_distance(point_1, point_2) -> float:
    """
    Compute Haversine distance (meters) between to points.
    Source: https://en.wikipedia.org/wiki/Haversine_formula

    Parameters
    ----------
    point_1 :
        First point.
    point_2 : _type_
        Second point.

    Returns
    -------
    float
        Haversine distance between the points.
    """
    # Delta and conversion to radians
    delta_lat = m.radians(point_1.lat - point_2.lat)
    delta_long = m.radians(point_1.lon - point_2.lon)

    sin_1 = m.sin(delta_lat / 2)
    sin_2 = m.sin(delta_long / 2)
    a = m.sqrt(
        sin_1 * sin_1
        + m.cos(m.radians(point_1.lat)) * m.cos(m.radians(point_2.lat)) * sin_2 * sin_2
    )
    d = 2 * EARTH_RADIUS * m.asin(a)

    return d


def distance(point_1, point_2) -> float:
    """
    Euclidian distance between two points.

    Parameters
    ----------
    point_1 :
        First point.
    point_2 :
        Second point.

    Returns
    -------
    float
        Distance between the points.
    """
    delta_lat = point_1.lat - point_2.lat
    delta_long = point_1.lon - point_2.lon
    return m.sqrt(delta_lat * delta_lat + delta_long * delta_long)


def perpendicular_distance(start_point, end_point, point) -> float:
    """
    Distance between a point and a line.

    Parameters
    ----------
    start_point :
        A point on the line.
    end_point :
        A point on the line.
    point :
        A point to measure the distance from.

    Returns
    -------
    float
        Perpendicular distance between the point *point* and the line defined
        by *start_point* and *end_point*.
    """

    def line_coefficients(point_1, point_2):
        """
        Compute the coefficients of a line equation of the form: ax+by+c=0.

        Parameters
        ----------
        point_1 :
            A point on the line.
        point_2 :
            A point on the line.

        Returns
        -------
        Tuple
            Coefficients of the line equation.
        """
        delta_x = point_1.lon - point_2.lon
        delta_y = point_1.lat - point_2.lat

        try:
            a = delta_y / delta_x
            b = -1
            c = point_1.lat - a * point_1.lon
        except:
            a = 1
            b = 0
            c = point_1.lon
            logging.debug("Vertical line")

        return a, b, c

    a, b, c = line_coefficients(start_point, end_point)

    d = abs(a * point.lon + b * point.lat + c) / m.sqrt(a * a + b * b)

    return d



import xml.etree.ElementTree as ET
from types import FunctionType
from typing import List


class GPXWriterMethodBehaviorCreator:
    def __init__(self):
        pass

    def add_bounds_creator(self, bounds_fields):
        code = (
            "def _add_bound(writer, element, bounds):"
            "\n\tif bounds is not None:"
            "\n\t\tbounds_ = ET.SubElement(element, bounds.tag)"
        )
        if "minlat" in bounds_fields:
            code += '\n\t\tbounds_, _ = writer.add_subelement_number(bounds_, "minlat", bounds.minlat, writer.precisions["lat_lon"])'
        if "minlon" in bounds_fields:
            code += '\n\t\tbounds_, _ = writer.add_subelement_number(bounds_, "minlon", bounds.minlon, writer.precisions["lat_lon"])'
        if "maxlat" in bounds_fields:
            code += '\n\t\tbounds_, _ = writer.add_subelement_number(bounds_, "maxlat", bounds.maxlat, writer.precisions["lat_lon"])'
        if "maxlon" in bounds_fields:
            code += '\n\t\tbounds_, _ = writer.add_subelement_number(bounds_, "maxlon", bounds.maxlon, writer.precisions["lat_lon"])'
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_bounds>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_bounds")
        return func

    def add_copyright_creator(self, copyright_fields):
        code = (
            "def _add_copyright(writer, element, copyright):"
            "\n\tif copyright is not None:"
            "\n\t\tcopyright_ = ET.SubElement(element, copyright.tag)"
        )
        if "author" in copyright_fields:
            code += '\n\t\twriter.setIfNotNone(copyright_, "author", copyright.author)'
        if "year" in copyright_fields:
            code += '\n\t\tcopyright_, _ = writer.add_subelement(copyright_, "year", str(copyright.year))'
        if "licence" in copyright_fields:
            code += '\n\t\tcopyright_, _ = writer.add_subelement(copyright_, "licence", str(copyright.licence))'
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_copyright>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_copyright")
        return func

    def add_email_creator(self, email_fields):
        code = (
            "def _add_email(writer, element, email):"
            "\n\tif email is not None:"
            "\n\t\temail_ = ET.SubElement(element, email.tag)"
        )
        if "id" in email_fields:
            code += '\n\t\twriter.setIfNotNone(email_, "id", email.id)'
        if "domain" in email_fields:
            code += '\n\t\twriter.setIfNotNone(email_, "domain", email.domain)'
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_email>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_email")
        return func

    # def add_extensions_creator(self, extensions_fields):
    #     code = ('def _add_extensions(writer, element, extensions):'
    #             '\n\tif extensions is not None:'
    #             '\n\t\textensions_ = ET.SubElement(element, extensions.tag)')
    #     if extensions_fields is not None:
    #         for k, v in extensions_fields:
    #             code += f'\n\t\textensions_, _ = writer.add_extensions_element(extensions_, "{k}", extensions.values["{field}"])'
    #     code += '\n\treturn element'
    #     compiled_code = compile(code, "<_add_extensions>", "exec")
    #     func = FunctionType(compiled_code.co_consts[0], globals(), "_add_extensions")
    #     return func

    def add_link_creator(self, link_fields):
        code = (
            "def _add_link(writer, element, link):"
            "\n\tif link is not None:"
            "\n\t\tlink_ = ET.SubElement(element, link.tag)"
        )
        if "href" in link_fields:
            code += (
                "\n\t\tif link.href is not None:"
                '\n\t\t\twriter.setIfNotNone(link_, "href", link.href)'
            )
        if "text" in link_fields:
            code += '\n\t\tlink_, _ = writer.add_subelement(link_, "text", link.text)'
        if "type" in link_fields:
            code += '\n\t\tlink_, _ = writer.add_subelement(link_, "type", link.type)'
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_link>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_link")
        return func

    def add_metadata_creator(self, metadata_fields: List[str]):
        code = (
            "def _add_metadata(writer, element, metadata):"
            "\n\tif metadata is not None:"
            "\n\t\tmetadata_ = ET.SubElement(element, metadata.tag)"
        )
        if "name" in metadata_fields:
            code += '\n\t\tmetadata_, _ = writer.add_subelement(metadata_, "name", metadata.name)'
        if "desc" in metadata_fields:
            code += '\n\t\tmetadata_, _ = writer.add_subelement(metadata_, "desc", metadata.desc)'
        if "author" in metadata_fields:
            code += "\n\t\tmetadata_ = writer.add_person(metadata_, metadata.author)"
        if "copyright" in metadata_fields:
            code += (
                "\n\t\tmetadata_ = writer.add_copyright(metadata_, metadata.copyright)"
            )
        if "link" in metadata_fields:
            code += "\n\t\tmetadata_ = writer.add_link(metadata_, metadata.link)"
        if "time" in metadata_fields:
            code += '\n\t\tmetadata_, _ = writer.add_subelement_time(metadata_, "time", metadata.time, writer.time_format)'
        if "keywords" in metadata_fields:
            code += '\n\t\tmetadata_, _ = writer.add_subelement(metadata_, "keywords", metadata.keywords)'
        if "bounds" in metadata_fields:
            code += "\n\t\tmetadata_ = writer.add_bounds(metadata_, metadata.bounds)"
        if "extensions" in metadata_fields:
            # code += '\n\t\tmetadata_ = writer.add_metadata_extensions(metadata_, metadata.extensions)'
            code += '\n\t\tmetadata_ = writer.add_extensions(metadata_, metadata.extensions, writer.extensions_fields.get("metadata"))'
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_metadata>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_metadata")
        return func

    def add_person_creator(self, person_fields):
        code = (
            "def _add_person(writer, element, person):"
            "\n\tif person is not None:"
            "\n\t\tperson_ = ET.SubElement(element, person.tag)"
        )
        if "name" in person_fields:
            code += (
                '\n\t\tperson_, _ = writer.add_subelement(person_, "name", person.name)'
            )
        if "email" in person_fields:
            code += "\n\t\tperson_ = writer.add_email(person_, person.email)"
        if "link" in person_fields:
            code += "\n\t\tperson_ = writer.add_link(person_, person.link)"
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_person>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_person")
        return func

    def add_point_segment_creator(self, point_segment_fields):
        code = (
            "def _add_point_segment(writer, element, point_segment):"
            "\n\tif point_segment is not None:"
            "\n\t\tpoint_segment_ = ET.SubElement(element, point_segment.tag)"
        )
        if "pt" in point_segment_fields:
            code += (
                "\n\t\tfor point in point_segment.pt:"
                "\n\t\t\tpoint_segment_ = writer.add_point(point_segment_, point)"
            )
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_point_segment>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_point_segment")
        return func

    def add_point_creator(self, point_fields):
        code = (
            "def _add_point(writer, element, point):"
            "\n\tif point is not None:"
            "\n\t\tpoint_ = ET.SubElement(element, point.tag)"
        )
        if "lat" in point_fields:
            code += '\n\t\twriter.setIfNotNone(point_, "lat", "{:.{}f}".format(point.lat, writer.precisions["lat_lon"]))'
        if "lon" in point_fields:
            code += '\n\t\twriter.setIfNotNone(point_, "lon", "{:.{}f}".format(point.lon, writer.precisions["lat_lon"]))'
        if "ele" in point_fields:
            code += '\n\t\tpoint_ = writer.add_subelement_number(point_, "ele", point.ele, writer.precisions["elevation"])'
        if "time" in point_fields:
            code += '\n\t\tpoint_, _ = writer.add_subelement_time(point_, "time", point.time, writer.time_format)'
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_point>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_point")
        return func

    def add_route_creator(self, route_fields):
        code = (
            "def _add_route(writer, element, route):"
            "\n\tif route is not None:"
            "\n\t\troute_ = ET.SubElement(element, route.tag)"
        )
        if "name" in route_fields:
            code += (
                '\n\t\troute_, _ = writer.add_subelement(route_, "name", route.name)'
            )
        if "cmt" in route_fields:
            code += '\n\t\troute_, _ = writer.add_subelement(route_, "cmt", route.cmt)'
        if "desc" in route_fields:
            code += (
                '\n\t\troute_, _ = writer.add_subelement(route_, "desc", route.desc)'
            )
        if "src" in route_fields:
            code += '\n\t\troute_, _ = writer.add_subelement(route_, "src", route.src)'
        if "link" in route_fields:
            code += "\n\t\troute_ = writer.add_link(route_, route.link)"
        if "number" in route_fields:
            code += '\n\t\troute_, _ = writer.add_subelement_number(route_, "number", route.src, 0)'
        if "type" in route_fields:
            code += (
                '\n\t\troute_, _ = writer.add_subelement(route_, "type", route.type)'
            )
        if "extensions" in route_fields:
            # code += '\n\t\troute_ = writer.add_rte_extensions(route_, route.extensions)'
            code += '\n\t\troute_ = writer.add_extensions(route_, route.extensions, writer.extensions_fields.get("rte"))'
        if "rtept" in route_fields:
            code += (
                "\n\t\tfor way_point in route.rtept:"
                "\n\t\t\troute_ = writer.add_way_point(route_, way_point)"
            )
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_route>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_route")
        return func

    def add_track_segment_creator(self, track_segment_fields):
        code = (
            "def _add_track_segment(writer, element, track_segment):"
            "\n\tif track_segment is not None:"
            "\n\t\ttrack_segment_ = ET.SubElement(element, track_segment.tag)"
        )
        if "extensions" in track_segment_fields:
            # code += '\n\t\ttrack_segment_ = writer.add_trkseg_extensions(track_segment_, track_segment.extensions)'
            code += '\n\t\ttrack_segment_ = writer.add_extensions(track_segment_, track_segment.extensions, writer.extensions_fields.get("trkseg"))'
        if "trkpt" in track_segment_fields:
            code += (
                "\n\t\tfor track_point in track_segment.trkpt:"
                "\n\t\t\ttrack_segment_ = writer.add_track_point(track_segment_, track_point)"
            )
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_track_segment>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_track_segment")
        return func

    def add_track_creator(self, track_fields):
        code = (
            "def _add_track(writer, element, track):"
            "\n\tif track is not None:"
            "\n\t\ttrack_ = ET.SubElement(element, track.tag)"
        )
        if "name" in track_fields:
            code += (
                '\n\t\ttrack_, _ = writer.add_subelement(track_, "name", track.name)'
            )
        if "cmt" in track_fields:
            code += '\n\t\ttrack_, _ = writer.add_subelement(track_, "cmt", track.cmt)'
        if "desc" in track_fields:
            code += (
                '\n\t\ttrack_, _ = writer.add_subelement(track_, "desc", track.desc)'
            )
        if "src" in track_fields:
            code += '\n\t\ttrack_, _ = writer.add_subelement(track_, "src", track.src)'
        if "link" in track_fields:
            code += "\n\t\ttrack_ = writer.add_link(track_, track.link)"
        if "number" in track_fields:
            code += '\n\t\ttrack_, _ = writer.add_subelement_number(track_, "number", track.src, 0)'
        if "type" in track_fields:
            code += (
                '\n\t\ttrack_, _ = writer.add_subelement(track_, "type", track.type)'
            )
        if "extensions" in track_fields:
            # code += '\n\t\ttrack_ = writer.add_trk_extensions(track_, track.extensions)'
            code += '\n\t\ttrack_ = writer.add_extensions(track_, track.extensions, writer.extensions_fields.get("trk"))'
        if "trkseg" in track_fields:
            code += (
                "\n\t\tfor track_seg in track.trkseg:"
                "\n\t\t\ttrack_ = writer.add_track_segment(track_, track_seg)"
            )
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_track>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_track")
        return func

    def add_way_point_creator(self, way_point_fields):
        code = (
            "def _add_way_point(writer, element, way_point):"
            "\n\tif way_point is not None:"
            "\n\t\tway_point_ = ET.SubElement(element, way_point.tag)"
        )
        if "lat" in way_point_fields:
            code += '\n\t\twriter.setIfNotNone(way_point_, "lat", "{:.{}f}".format(way_point.lat, writer.precisions["lat_lon"]))'
        if "lon" in way_point_fields:
            code += '\n\t\twriter.setIfNotNone(way_point_, "lon", "{:.{}f}".format(way_point.lon, writer.precisions["lat_lon"]))'
        if "ele" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "ele", way_point.ele, writer.precisions["elevation"])'
        if "time" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_time(way_point_, "time", way_point.time, writer.time_format)'
        if "magvar" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "magvar", way_point.mag_var, writer.precisions["default"])'
        if "geoidheight" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "geoidheight", way_point.geo_id_height, writer.precisions["default"])'
        if "name" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "name", way_point.name)'
        if "cmt" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "cmt", way_point.cmt)'
        if "desc" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "desc", way_point.desc)'
        if "src" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "src", way_point.src)'
        if "link" in way_point_fields:
            code += "\n\t\tway_point_ = writer.add_link(way_point_, way_point.link)"
        if "sym" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "sym", way_point.sym)'
        if "type" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "type", way_point.type)'
        if "fix" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "fix", way_point.fix)'
        if "sat" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "sat", way_point.sat, 0)'
        if "hdop" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "hdop", way_point.hdop, writer.precisions["default"])'
        if "vdop" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "vdop", way_point.vdop, writer.precisions["default"])'
        if "pdop" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "pdop", way_point.pdop, writer.precisions["default"])'
        if "ageofgpsdata" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "ageofgpsdata", way_point.age_of_gps_data, writer.precisions["default"])'
        if "dgpsid" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "dgpsid", way_point.dgpsid, 0)'
        if "extensions" in way_point_fields:
            # code += '\n\t\tway_point_ = writer.add_wpt_extensions(way_point_, way_point.extensions)'
            code += '\n\t\tway_point_ = writer.add_extensions(way_point_, way_point.extensions, writer.extensions_fields.get("wpt"))'
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_way_point>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_way_point")
        return func

    def add_track_point_creator(self, way_point_fields):
        code = (
            "def _add_track_point(writer, element, way_point):"
            "\n\tif way_point is not None:"
            "\n\t\tway_point_ = ET.SubElement(element, way_point.tag)"
        )
        if "lat" in way_point_fields:
            code += '\n\t\twriter.setIfNotNone(way_point_, "lat", "{:.{}f}".format(way_point.lat, writer.precisions["lat_lon"]))'
        if "lon" in way_point_fields:
            code += '\n\t\twriter.setIfNotNone(way_point_, "lon", "{:.{}f}".format(way_point.lon, writer.precisions["lat_lon"]))'
        if "ele" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "ele", way_point.ele, writer.precisions["elevation"])'
        if "time" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_time(way_point_, "time", way_point.time, writer.time_format)'
        if "magvar" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "magvar", way_point.mag_var, writer.precisions["default"])'
        if "geoidheight" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "geoidheight", way_point.geo_id_height, writer.precisions["default"])'
        if "name" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "name", way_point.name)'
        if "cmt" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "cmt", way_point.cmt)'
        if "desc" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "desc", way_point.desc)'
        if "src" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "src", way_point.src)'
        if "link" in way_point_fields:
            code += "\n\t\tway_point_ = writer.add_link(way_point_, way_point.link)"
        if "sym" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "sym", way_point.sym)'
        if "type" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "type", way_point.type)'
        if "fix" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement(way_point_, "fix", way_point.fix)'
        if "sat" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "sat", way_point.sat, 0)'
        if "hdop" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "hdop", way_point.hdop, writer.precisions["default"])'
        if "vdop" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "vdop", way_point.vdop, writer.precisions["default"])'
        if "pdop" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "pdop", way_point.pdop, writer.precisions["default"])'
        if "ageofgpsdata" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "ageofgpsdata", way_point.age_of_gps_data, writer.precisions["default"])'
        if "dgpsid" in way_point_fields:
            code += '\n\t\tway_point_, _ = writer.add_subelement_number(way_point_, "dgpsid", way_point.dgpsid, 0)'
        if "extensions" in way_point_fields:
            # code += '\n\t\tway_point_ = writer.add_trkpt_extensions(way_point_, way_point.extensions)'
            code += '\n\t\tway_point_ = writer.add_extensions(way_point_, way_point.extensions, writer.extensions_fields.get("trkpt"))'
        code += "\n\treturn element"
        compiled_code = compile(code, "<_add_track_point>", "exec")
        func = FunctionType(compiled_code.co_consts[0], globals(), "_add_track_point")
        return func



import datetime

from .gpx_element import GpxElement


class Point(GpxElement):
    """
    ptType element in GPX file.
    """

    fields = ["lat", "lon", "ele", "time"]
    mandatory_fields = ["lat", "lon"]

    def __init__(
        self,
        tag: str = "pt",
        lat: float = None,
        lon: float = None,
        ele: float = None,
        time: datetime = None,
    ) -> None:
        """
        Initialize Point instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "pt"
        lat : float, optional
            Latitude, by default None
        lon : float, optional
            Longitude, by default None
        ele : float, optional
            Elevation, by default None
        time : datetime, optional
            Time, by default None
        """
        self.tag: str = tag
        self.lat: float = lat
        self.lon: float = lon
        self.ele: float = ele
        self.time: datetime = time



import pandas as pd

from ..gpx import GPX


class Plotter:
    """
    GPX plotter (parent class).
    """

    def __init__(self, gpx: GPX) -> None:
        """
        Initialise base class plotter instance.

        Args:
            gpx (GPX): GPX object to plot.
        """
        self._gpx: GPX = gpx

        # self._background: str = None
        # self._color: str = "#FFA800"
        # self._start_color: str = "#00FF00"
        # self._stop_color: str = "#FF0000"
        # self._way_points_color: str = "#0000FF"
        # self._title: str = None
        # self._file_path: str = None

        self._dataframe: pd.DataFrame = None

    def plot(self):
        """
        Plot GPX.
        """



import os
import warnings
from math import isclose
from typing import Optional, Tuple

import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap

from .plotter import Plotter


class MatplotlibPlotter(Plotter):
    """
    GPX plotter based on Matplotlib.
    """

    def plot(
        self,
        figsize: Tuple[int, int] = (16, 9),
        size: float = 5,
        color: str = "#FFA800",
        cmap: Optional[matplotlib.colors.Colormap] = None,
        colorbar: bool = False,
        start_point_color: Optional[str] = None,
        stop_point_color: Optional[str] = None,
        way_points_color: Optional[str] = None,
        background: Optional[str] = None,
        offset_percentage: float = 0.04,
        dpi: int = 96,
        title: Optional[str] = None,
        title_fontsize: int = 20,
        watermark: bool = False,
        file_path: str = None,
    ):
        """
        Plot GPX using Matplotlib.

        Args:
            figsize (Tuple[int, int], optional): Width and height of
                the plot. Defaults to (16, 9).
            size (float, optional): Size of the track. Defaults to 5.
            color (str, optional): Color of the track. Possible choice
                are: "lat", "lon", "ele", "speed", "pace",
                "vertical_drop", "ascent_rate", "ascent_speed" or any
                valid color string. Defaults to "#FFA800".
            cmap (Optional[matplotlib.colors.Colormap], optional):
                Colormap to use when color is not a fixed color value.
                Defaults to None.
            colorbar (bool, optional): Add colorbar. Defaults to False.
            start_point_color (Optional[str], optional): Color of the
                first point. Defaults to None.
            stop_point_color (Optional[str], optional): Color of the
                last point. Defaults to None.
            way_points_color (Optional[str], optional): Color of the
                way points. Defaults to None.
            background (Optional[str], optional): Map tiles to use.
                Possible choice are: None, "bluemarble",
                "shadedrelief", "etopo", "World_Imagery", "wms" or any
                server supported by
                `mpl_toolkits.basemap.Basemap.arcgisimage`. Defaults to
                None.
            offset_percentage (float, optional): Offset percentage to
                apply to the track bounding box. Defaults to 0.04.
            dpi (int, optional): Resolution of the animation. Defaults
                to 96.
            interval (float, optional): Interval between frames of the
                animation. Defaults to 20.
            bitrate (int, optional): Bit-rate of the animation.
                Defaults to 1800.
            repeat (bool, optional): Repeat the animation when viewed.
                Defaults to True.
            title (Optional[str], optional): Title of the plot.
                Defaults to None.
            title_fontsize (int, optional): Font size of the title of
                the plot. Defaults to 20.
            watermark (bool, optional): Watermark. Defaults to False.
            file_path (str, optional): Path to save the plot. Defaults
                to None.

        Raises:
            FileNotFoundError: Provided path does not exist.

        Returns:
            matplotlib.Figure: Plot of the GPX.
        """
        dynamic_colors = [
            "lat",
            "lon",
            "ele",
            "speed",
            "pace",
            "vertical_drop",
            "ascent_rate",
            "ascent_speed",
        ]

        # Create dataframe containing data from the GPX file
        values = ["lat", "lon"]
        if color in dynamic_colors:
            values.append(color)
        self._dataframe = self._gpx.to_pandas(values)

        # Create figure
        fig = plt.figure(figsize=figsize)

        # Compute track boundaries
        min_lat, min_lon, max_lat, max_lon = self._gpx.bounds()

        # Compute default offset
        delta_lat = abs(max_lat - min_lat)
        delta_lon = abs(max_lon - min_lon)
        delta_max = max(delta_lat, delta_lon)
        offset = delta_max * offset_percentage

        # Add default offset
        min_lat = max(-90, min_lat - offset)
        max_lat = min(max_lat + offset, 90)
        min_lon = max(-180, min_lon - offset)
        max_lon = min(max_lon + offset, 180)

        # Update min/max lat/lon to achieve correct aspect ratio
        lat_offset = 1e-5
        lon_offset = 1e-5
        delta_lat = abs(max_lat - min_lat)
        delta_lon = abs(max_lon - min_lon)
        r = delta_lon / delta_lat  # Current map aspect ratio
        r_ref = figsize[0] / figsize[1]  # Target map aspect ratio
        tolerance = 1e-3
        while not isclose(r, r_ref, abs_tol=tolerance):
            if r > r_ref:
                min_lat = max(-90, min_lat - lat_offset)
                max_lat = min(max_lat + lat_offset, 90)
                delta_lat = max_lat - min_lat
            if r < r_ref:
                min_lon = max(-180, min_lon - lon_offset)
                max_lon = min(max_lon + lon_offset, 180)
                delta_lon = max_lon - min_lon
            r = delta_lon / delta_lat

        # Create map
        map_ = Basemap(
            projection="cyl",
            llcrnrlon=min_lon,
            llcrnrlat=min_lat,
            urcrnrlon=max_lon,
            urcrnrlat=max_lat,
        )

        # Add background
        if background is None:
            pass
        elif background == "bluemarble":
            map_.bluemarble()
        elif background == "shadedrelief":
            map_.shadedrelief()
        elif background == "etopo":
            map_.etopo()
        elif background == "World_Imagery":
            map_.arcgisimage(service=background, dpi=dpi)
        elif background == "wms":
            wms_server = "http://www.ga.gov.au/gis/services/topography/Australian_Topography/MapServer/WMSServer"
            wms_server = "http://wms.geosignal.fr/metropole?"
            map_.wmsimage(
                wms_server, layers=["Communes", "Nationales", "Regions"], verbose=True
            )
        else:
            map_.arcgisimage(service=background, dpi=dpi, verbose=True)

        # Scatter track points
        if color in dynamic_colors:
            im = map_.scatter(
                self._dataframe["lon"],
                self._dataframe["lat"],
                s=size,
                c=self._dataframe[color],
                cmap=cmap,
            )
        else:
            im = map_.scatter(
                self._dataframe["lon"], self._dataframe["lat"], s=size, color=color
            )

        # Scatter start point with different color
        if start_point_color:
            map_.scatter(
                self._dataframe["lon"][0],
                self._dataframe["lat"][0],
                marker="^",
                color=start_point_color,
            )

        # Scatter stop point with different color
        if stop_point_color:
            map_.scatter(
                self._dataframe["lon"].iloc[-1],
                self._dataframe["lat"].iloc[-1],
                marker="h",
                color=stop_point_color,
            )

        # Scatter way points with different color
        if way_points_color:
            for way_point in self._gpx.wpt:
                x, y = map_(way_point.lon, way_point.lat)  # Project way point
                map_.scatter(
                    x, y, marker="D", color=way_points_color
                )  # Scatter way point

        # Colorbar
        if colorbar:
            fig.colorbar(im)

        # Add title
        if title is not None:
            if watermark:
                fig.suptitle(title + "\n[made with ezGPX]", fontsize=title_fontsize)
            else:
                fig.suptitle(title, fontsize=title_fontsize)

        # Set figure layout
        fig.tight_layout()

        # Save plot
        if file_path is not None:
            # Check if provided path exists
            directory_path = os.path.dirname(os.path.realpath(file_path))
            if not os.path.exists(directory_path):
                raise FileNotFoundError("Provided path does not exist")
            fig.savefig(file_path)

        return fig



import logging
import os
from typing import Optional

import plotly.graph_objects as go

from .plotter import Plotter


class PlotlyPlotter(Plotter):
    """
    GPX plotter based on Plotly.
    """

    def plot(
        self,
        tiles: str = "open-street-map",  # "open-street-map"
        mode: str = "lines",  # "lines", "markers+lines"
        color: str = "#FFA800",
        start_point_color: Optional[str] = None,
        stop_point_color: Optional[str] = None,
        way_points_color: Optional[str] = None,
        title: Optional[str] = None,
        zoom: float = 12.0,
        file_path: Optional[str] = None,
    ):
        """
        Plot (animation) GPX using Plotly.

        Args:
            tiles (str, optional): Map tiles to use. Defaults to
                "open-street-map".
            start_point_color (Optional[str], optional): Color of the
                first point. Defaults to None.
            stop_point_color (Optional[str], optional): Color of the
                last point. Defaults to None.
            way_points_color (Optional[str], optional): Color of the
                way points. Defaults to None.
            title (Optional[str], optional): Title of the plot.
                Defaults to None.
            zoom (float, optional): Zoom. Defaults to 12.0.
            file_path (Optional[str], optional): Path to save the plot. Defaults to None.

        Returns:
            _type_: _description_
        """
        self._dataframe = self._gpx.to_pandas()
        center_lat, center_lon = self._gpx.center()

        # Create map and scatter
        fig = go.Figure(
            go.Scattermap(
                mode=mode,
                lon=self._dataframe["lon"],
                lat=self._dataframe["lat"],
                marker={"size": 5, "color": color},
            )
        )

        # Scatter start and stop points with different color
        if start_point_color:
            fig.add_trace(
                go.Scattermap(
                    mode="markers",
                    lon=[self._dataframe["lon"].iloc[0]],
                    lat=[self._dataframe["lat"].iloc[0]],
                    marker={"size": 5, "color": start_point_color},
                )
            )
        if stop_point_color:
            fig.add_trace(
                go.Scattermap(
                    mode="markers",
                    lon=[self._dataframe["lon"].iloc[-1]],
                    lat=[self._dataframe["lat"].iloc[-1]],
                    marker={"size": 5, "color": stop_point_color},
                )
            )

        # Scatter way points with different color
        if way_points_color:
            for way_point in self._gpx.gpx.wpt:
                fig.add_trace(
                    go.Scattermap(
                        mode="markers",
                        lon=way_point.lon,
                        lat=way_point.lat,
                        marker={"size": 5, "color": way_points_color},
                    )
                )

        # Update layout for nice display
        fig.update_layout(
            margin={"l": 0, "t": 0, "b": 0, "r": 0},
            map={
                "center": {"lon": center_lon, "lat": center_lat},
                "style": tiles,
                "zoom": zoom,
            },
            title={"text": title},
        )

        # Save plot
        if file_path is not None:
            # Check if provided path exists
            directory_path = os.path.dirname(os.path.realpath(file_path))
            if not os.path.exists(directory_path):
                logging.error("Provided path does not exist")
                return
            if file_path.endswith(".html"):
                fig.write_html(file_path)
            elif file_path.endswith(".json"):
                fig.write_json(file_path)
            else:
                fig.write_image(file_path)

        return fig



from .email import Email
from .gpx_element import GpxElement
from .link import Link


class Person(GpxElement):
    """
    personType element in GPX file.
    """

    fields = ["name", "email", "link"]
    mandatory_fields = []

    def __init__(
        self,
        tag: str = "person",
        name: str = None,
        email: Email = None,
        link: Link = None,
    ) -> None:
        """
        Initialize Person instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "person"
        name : str, optional
            Name, by default None
        email : Email, optional
            Email, by default None
        link : Link, optional
            Link, by default None
        """
        self.tag: str = tag
        self.name: str = name
        self.email: Email = email
        self.link: Link = link



from .gpx_element import GpxElement


class Extensions(GpxElement):
    """
    extensionsType element in GPX file.
    """

    fields = []
    mandatory_fields = []

    def __init__(self, tag: str = "extensions", values: dict = None):
        self.tag: str = tag
        self.values: dict = values if values is not None else {}



from typing import List

from .gpx_element import GpxElement
from .point import Point


class PointSegment(GpxElement):
    """
    ptsegType element in GPX file.
    """

    fields = ["pt"]
    mandatory_fields = []

    def __init__(self, tag: str = "ptseg", pt: List[Point] = None) -> None:
        """
        Initialize PointSegment instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "ptseg"
        pt : List[Point], optional
            List of points, by default None
        """
        self.tag: str = tag
        self.pt: List[Point] = [] if pt is None else pt



import logging
import xml.etree.ElementTree as ET
from datetime import datetime
from typing import Dict, Optional, Union

from dateutil import parser

from .parser import Parser


class XMLParser(Parser):
    """
    XML File parser.
    """

    def __init__(
        self,
        file_path: Optional[str] = None,
        xml_schema: bool = True,
        xml_extensions_schemas: bool = False,
    ) -> None:
        """
        Initialize XML Parser instance.

        Parameters
        ----------
        file_path : Optional[str], optional
            Path to the file to parse, by default None
        xml_schema : bool, optional
            Toggle  schema verification during parsing, by default True
        xml_extensions_schemas : bool, optional
            Toggle extensions schema verificaton durign parsing.
            Requires internet connection and is not guaranted to work,
            by default False
        """
        self.name_spaces: dict = dict(
            [node for _, node in ET.iterparse(file_path, events=["start-ns"])]
        )
        self.extensions_fields: Dict = {}

        super().__init__(file_path, self.name_spaces)

        self.xml_schema: bool = xml_schema
        self.xml_extensions_schemas: bool = xml_extensions_schemas

        self.xml_tree: ET.ElementTree = None
        self.xml_root: ET.Element = None

    def get_text(self, element, sub_element: str) -> Union[str, None]:
        """
        Get text from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[str, None]: Text from sub-element.
        """
        text_ = element.get(sub_element)
        if text_ is None:
            logging.debug("%s has no attribute %s.", element, sub_element)
        return text_

    def get_int(self, element, sub_element: str) -> Union[int, None]:
        """
        Get integer value from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[int, None]: Integer value from sub-element.
        """
        int_ = element.get(sub_element)
        if int_ is None:
            logging.debug("%s has no attribute %s.", element, sub_element)
        else:
            int_ = int(int_)
        return int_

    def get_float(self, element, sub_element: str) -> Union[float, None]:
        """
        Get floating point value from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[float, None]: Floating point value from sub-element.
        """
        float_ = element.get(sub_element)
        if float_ is None:
            logging.debug("%s has no attribute %s.", element, sub_element)
        else:
            float_ = float(float_)
        return float_

    def find_sub_element(self, element, sub_element: str) -> Union[ET.Element, None]:
        """
        Find sub-element.

        Parameters
        ----------
        element : xml.etree.ElementTree.Element
            Parsed element from GPX file.
        sub_element : str
            Sub-element name.

        Returns
        -------
        Union[ET.Element, None]
            Sub-element.
        """
        sub_element_ = element.find(sub_element, self.name_spaces)
        if sub_element_ is None:
            logging.debug("%s has no attribute %s.", element, sub_element)
        return sub_element_

    def find_text(self, element, sub_element: str) -> Union[str, None]:
        """
        Find text from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[str, None]: Text from sub-element.
        """
        sub_element_ = self.find_sub_element(element, sub_element)
        return None if sub_element_ is None else sub_element_.text

    def find_int(self, element, sub_element: str) -> Union[int, None]:
        """
        Find integer value from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[int, None]: Integer value from sub-element.
        """
        sub_element_ = self.find_sub_element(element, sub_element)
        return None if sub_element_ is None else int(sub_element_.text)

    def find_float(self, element, sub_element: str) -> Union[float, None]:
        """
        Find float point value from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[float, None]: Floating point value from sub-element.
        """
        sub_element_ = self.find_sub_element(element, sub_element)
        return None if sub_element_ is None else float(sub_element_.text)

    def find_time(self, element, sub_element: str) -> Union[datetime, None]:
        """
        Find time value from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[datetime, None]: Floating point value from sub-element.
        """
        sub_element_ = self.find_sub_element(element, sub_element)
        return None if sub_element_ is None else parser.parse(sub_element_.text)

    def check_xml_schemas(self):
        """
        Check XML schemas during parsing.
        """
        # Check XML schema
        if self.xml_schema:
            if not self.gpx.check_xml_schema(self.file_path):
                raise ValueError("Invalid GPX file (does not follow XML schema).")

        # Check XML extension schemas
        if self.xml_extensions_schemas:
            if not self.gpx.check_xml_extensions_schemas(self.file_path):
                raise ValueError(
                    "Invalid GPX file (does not follow XML extensions schemas)."
                )
